#!/usr/bin/env python
#
# MELDREW, the IRC game interpreter
# http://www.icculus.org/projects/meldrew/
#
# (c) 2001 Michael S. Phillips
# msphil@icculus.org
#
# This IRC bot allows multiple persons to play any supported game type
# (requires a Glk-based interpreter, and a game playable with CheapGlk)
# via /msg
#
# Multiple persons may play simultaneously.
#
# Game files, short names, descriptions, and interpreters are stored in 
# meldrew.rc in the meldrew directory.  The need-to-configure options are 
# just below.
#
# Full documentation, design, hooks, etc. is in the meldrew.txt file that 
# accompanies this program.
#
# See the file ChangeLog for changes from version to version, and TODO 
# for the current TODO list.
# 
#-----------------------------------------------------------------------
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#        
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#-----------------------------------------------------------------------

import sys
import os
import string
import anydbm
import thread
import time

from ircbot import SingleServerIRCBot
from irclib import nm_to_n, irc_lower

meldrew_version = "0.1.1"

# defaults: these should be in meldrew.rc
default_game_name = "curses"
default_irc_name = "meldrew"
default_irc_channel = "#loki"
default_irc_host = "irc.openprojects.net"
default_max_concurrent = 16  #maximum concurrent terp sessions
default_max_lines = 8        #after 8 lines, throttle the output queue
default_terp_timeout = 30    #inactivity in minutes, after which the terp is killed
default_superadmin = "meladmin"

meldrew_admin = "meladmin"

# a couple of boring globals:
terp_list = []
game_list = []
connected_users = {}
currently_playing = {}
user_db = anydbm.open( "meldrew.db", 'c' )

# A couple of handy utility functions
# load config

class MeldrewBot(SingleServerIRCBot):
    def __init__(self, channel, nickname, server, port=6667):
        SingleServerIRCBot.__init__(self, [(server, port)], nickname, nickname)
        self.channel = channel

        self.children_in = {}
        self.children_out = {}
        self.is_playing = {}

        self.start()

    def on_welcome(self, c, e):
        c.join(self.channel)

    def on_privmsg( self, c, e):
        self.do_command(nm_to_n(e.source()), e.arguments()[0])

    def on_pubmsg( self, c, e):
        a = string.split(e.arguments()[0], ":", 1)
        if len(a) > 1 and irc_lower(a[0]) == irc_lower(self.connection.get_nickname()):
            self.do_command(nm_to_n(e.source()), string.strip(a[1]))
        return

    def run_terp(self, nick ):
        self.is_playing[nick] = 1
        while 1:
            line = self.children_out[nick].readline();
            if line == None:
                print "We's outta here! (None)"
                break
            if len(line) == 0:
                print "We's outta here! (len 0)"
                break
            line = line[:-1]
            if len(line) > 2:
                self.privmsg( nick, line )
	        time.sleep(1)
        self.children_out[nick].close()
        self.children_in[nick].close()
        self.privmsg( nick, "The interpreter has ended the session." )
        self.is_playing[nick] = 0
        thread.exit()

    def do_command(self, nick, cmd):
        print ">%s<: %s" % (nick, cmd)
        c = self.connection

        inputline = cmd

#-----------------------------------------------------------------------
# Check to see if user is currently playing
#-----------------------------------------------------------------------

        if self.is_playing.has_key(nick):
	    if self.is_playing[nick] == 1:
	        self.children_in[nick].write( "%s\n" % cmd )
	        self.children_in[nick].flush()
	        return

#-----------------------------------------------------------------------
# !quit
#-----------------------------------------------------------------------
        if inputline[:5] == "!quit":
            if connected_users.has_key(nick):
                if connected_users[nick] == meldrew_admin:
                    self.privmsg( nick, "OK, boss!" )
                    self.die()
                else:
                    self.privmsg( nick, "Sorry, Meldrew doesn't recognize your authority." )
            else:
                self.privmsg( nick, "Sorry, Meldrew doesn't recognize you to determine your authority." )

#-----------------------------------------------------------------------
# !how[to]
#-----------------------------------------------------------------------
        elif inputline[:4] == "!how":
            self.privmsg( nick, "How do I use this bot?" )
            self.privmsg( nick, " " )
            self.privmsg( nick, "First, you need to register a userid, using REGISTER:" )
            self.privmsg( nick, "    !register thomas jefferson" )
            self.privmsg( nick, "(Your userid does NOT need to be your nick, and it DOES need to be " )
            self.privmsg( nick, "a single word.)" )
            self.privmsg( nick, " " )
            self.privmsg( nick, "Then, you need to identify yourself, using IDENTIFY:" )
            self.privmsg( nick, "    !identify thomas jefferson" )
            self.privmsg( nick, "You may then play any supported game, using /msg with the bot.  A" )
            self.privmsg( nick, "save game directory is created on the host for each game and userid," )
            self.privmsg( nick, "so please be kind and use !clear to remove your directory when you're" )
            self.privmsg( nick, "done with a specific game." )
            self.privmsg( nick, " " )
            self.privmsg( nick, "Note that the interpreters have some limitations (no terminal), and " )
            self.privmsg( nick, "the menus may be difficult to navigate.  However, most games have " )
            self.privmsg( nick, "both general IF instructions and specific instructions for the " )
            self.privmsg( nick, "particular game." )
            self.privmsg( nick, " " )
            self.privmsg( nick, "Meldrew tries to be kind to the servers, so he will slow down his" )
            self.privmsg( nick, "output if it gets over a certain threshold.  (You may have already" )
            self.privmsg( nick, "noticed this with the longer !command messages, such as this one...)" )
            self.privmsg( nick, " " )
            self.privmsg( nick, "If you decide you will not be using the bot ever again, please be " )
            self.privmsg( nick, "kind and do your !unregister:" )
            self.privmsg( nick, "    !unregister thomas jefferson" )
            self.privmsg( nick, "This will remove all data associated with the userid." )

#-----------------------------------------------------------------------
# !help
#-----------------------------------------------------------------------
        elif inputline[:5] == "!help":
            self.privmsg( nick, "    !list                         -- lists the games to play" )
            self.privmsg( nick, "    !list [type]                  -- lists the games of a specific type" )
            self.privmsg( nick, "    !terp                         -- lists known interpreters" )
            self.privmsg( nick, "    !play [game]                  -- plays game (e.g. !play curses)" )
            self.privmsg( nick, " " )
            self.privmsg( nick, "    !register [userid] [passwd]   -- creates [userid] with [passwd]" )
            self.privmsg( nick, "    !unregister [userid] [passwd] -- destroys [userid] with [passwd]," )
            self.privmsg( nick, "                                     including savegame directories" )
            self.privmsg( nick, "    !reregister [userid] [oldpass] [newpass]" )
            self.privmsg( nick, "                                  -- changes password on [userid]" )
            self.privmsg( nick, "    !identify [userid] [passwd]   -- start a session as [userid]" )
            self.privmsg( nick, "    !unidentify                   -- end your session" )
            self.privmsg( nick, "    !saves                        -- after you have identified yourself," )
            self.privmsg( nick, "                                     shows your current save game groups" )
            self.privmsg( nick, "    !saves [game]                 -- shows the saves in a particular game" )
            self.privmsg( nick, "    !clear [game]                 -- clears the saved game files for a " )
            self.privmsg( nick, "                                     particular game" )
            self.privmsg( nick, " " )
            self.privmsg( nick, "    !quit                         -- quits if being run interactively" )
            self.privmsg( nick, "    !info                         -- information about this bot" )
            self.privmsg( nick, "    !howto                        -- how to use this bot" )
            self.privmsg( nick, "    !numclients                   -- how many players, etc." )
            self.privmsg( nick, "    !version                      -- displays Meldrew's version" )
            self.privmsg( nick, "    !help                         -- shows this message" )
#-----------------------------------------------------------------------
# !version
#-----------------------------------------------------------------------
        elif (inputline[:8] == "!version"):
            self.privmsg( nick, "Meldrew" )
            self.privmsg( nick, meldrew_version )

#-----------------------------------------------------------------------
# !list
#-----------------------------------------------------------------------
        elif (inputline[:5] == "!list"):
            list_list = string.splitfields( inputline, " " )
            if ( len(list_list) > 1 ):
                game_count = 0;
                for i in range(len(game_list)):
                    if game_list[i][3] == list_list[1]:
                        self.privmsg( nick, '%s: %s' % (game_list[i][0], game_list[i][2]) )
			game_count = game_count + 1
		self.privmsg( nick, '%d games listed' % game_count )
            else:
                for i in range(len(game_list)):
                    self.privmsg( nick, '%s[%s]: %s' % (game_list[i][0], game_list[i][3], game_list[i][2]) )
		self.privmsg( nick, '%d games listed' % len(game_list) )

#-----------------------------------------------------------------------
# !game
#-----------------------------------------------------------------------
        elif (inputline[:5] == "!game"):
            for i in range(len(game_list)):
                self.privmsg( nick, '%s[%s]: %s' % (game_list[i][0], game_list[i][3], game_list[i][2]) )

#-----------------------------------------------------------------------
# !terp
#-----------------------------------------------------------------------
        elif inputline[:5] == "!terp":
            for i in range(len(terp_list)):
                self.privmsg( nick, '%s[%s]' % (terp_list[i][1], terp_list[i][0]) )

#-----------------------------------------------------------------------
# !register
#-----------------------------------------------------------------------
        elif inputline[:9] == "!register":
            user_list = []
            user_list_expanded = string.splitfields( inputline, " " )
            for i in range(len(user_list_expanded)):
                if user_list_expanded[i] != '':
                    user_list.append( user_list_expanded[i] )
            if len(user_list) != 3:
                self.privmsg( nick, "Meldrew expects the format '!register [userid] [password]'" )
            elif user_db.has_key( user_list[1] ):
                self.privmsg( nick, "Sorry, Meldrew already knows that userid." )
            else:
                self.privmsg( nick, "Adding %s" % user_list[1] )
                self.privmsg( nick, "Meldrew hopes you remember your password!" )
		os.makedirs( "saves/%s" % user_list[1] )
                user_db[user_list[1]] = user_list[2]

#-----------------------------------------------------------------------
# !unregister
#-----------------------------------------------------------------------
        elif inputline[:11] == "!unregister":
            user_list = []
            user_list_expanded = string.splitfields( inputline, " " )
            for i in range(len(user_list_expanded)):
                if user_list_expanded[i] != '':
                    user_list.append( user_list_expanded[i] )
            if len(user_list) != 3:
                self.privmsg( nick, "Meldrew expects the format '!unregister [userid] [password]'" )
            elif user_db.has_key( user_list[1] ):
                if user_list[2] == user_db[user_list[1]]:
                    self.privmsg( nick, "Deleting %s" % user_list[1] )
                    self.privmsg( nick, "Meldrew hopes you don't need it any more!" )
                    del user_db[user_list[1]]
                else:
                    self.privmsg( nick, "Sorry, Meldrew believes that is the wrong password for this userid" )
            else:
                self.privmsg( nick, "Sorry, Meldrew doesn't know that userid." )


#-----------------------------------------------------------------------
# !reregister
#-----------------------------------------------------------------------
        elif inputline[:11] == "!reregister":
            user_list = []
            user_list_expanded = string.splitfields( inputline, " " )
            for i in range(len(user_list_expanded)):
                if user_list_expanded[i] != '':
                    user_list.append( user_list_expanded[i] )
            if len(user_list) != 4:
                self.privmsg( nick, "Meldrew expects the format '!reregister [userid] [oldpass] [newpass]'" )
            elif user_db.has_key( user_list[1] ):
                if user_list[2] == user_db[user_list[1]]:
                    self.privmsg( nick, "Changing password for %s" % user_list[1] )
                    self.privmsg( nick, "Meldrew hopes you remember the new password!" )
                    user_db[user_list[1]] = user_list[3]
                else:
                    self.privmsg( nick, "Sorry, Meldrew believes that is the wrong password for this userid" )
            else:
                self.privmsg( nick, "Sorry, Meldrew doesn't know that userid." )

#-----------------------------------------------------------------------
# !identify
#-----------------------------------------------------------------------
        elif inputline[:9] == "!identify":
            user_list = []
            user_list_expanded = string.splitfields( inputline, " " )
            for i in range(len(user_list_expanded)):
                if user_list_expanded[i] != '':
                    user_list.append( user_list_expanded[i] )
            if len(user_list) != 3:
                self.privmsg( nick, "Meldrew expects the format '!identify [userid] [password]'" )
            elif user_db.has_key( user_list[1] ):
                if user_list[2] == user_db[user_list[1]]:
                    self.privmsg( nick, 'Meldrew now recognizes you as %s' % user_list[1] )
                    connected_users[nick] = user_list[1]
                else:
                    self.privmsg( nick, "Meldrew believes that is the wrong password for this userid" )

#-----------------------------------------------------------------------
# !unidentify
#-----------------------------------------------------------------------
        elif inputline[:11] == "!unidentify":
            if connected_users.has_key(nick):
                self.privmsg( nick, "Meldrew has forgotten that you are %s" % connected_users[nick] )
                del connected_users[nick]
            else:
                self.privmsg( nick, "Meldrew doesn't recognize you" )

#-----------------------------------------------------------------------
# !play
#-----------------------------------------------------------------------
        elif inputline[:5] == "!play":
            if connected_users.has_key(nick):
                shortname = inputline[6:]
                if shortname == "":
                    shortname = default_game_name;
                game_to_play = -1;
                for i in range(len(game_list)):
                    if ( game_list[i][0] == shortname ):
                        game_to_play = i;
                if game_to_play != -1:
                    terp_to_use = -1;
                    for i in range(len(terp_list)):
                        if ( terp_list[i][0] == game_list[game_to_play][3] ):
                            terp_to_use = i
                    if terp_to_use != -1:
                        terp_cmd = "%s %s %s" % (terp_list[terp_to_use][2], terp_list[terp_to_use][3], game_list[game_to_play][1])
                        print "PLAYING: %s" % terp_cmd
			try:
			    os.makedirs( "saves/%s/%s" % (connected_users[nick], shortname ) )
			except:
			    #
			    pass
			os.chdir( "saves/%s/%s" % (connected_users[nick], shortname) )
                        (self.children_in[nick], self.children_out[nick]) = os.popen2( terp_cmd )
			os.chdir( "../../.." )
                        thread.start_new_thread( self.run_terp, (nick,) )
                    else:
                        self.privmsg( nick, "Sorry, sir, Meldrew doesn't have a terp for that!" )
                else:
                    self.privmsg( nick, "Sorry, sir, Meldrew don't play dat!" )
	    else:
	        self.privmsg( nick, "Sorry, sir, Meldrew requires you to identify yourself first." )

#-----------------------------------------------------------------------
# !numclients
#-----------------------------------------------------------------------
        elif inputline[:11] == "!numclients":
            if len(connected_users) == 1:
                self.privmsg( nick, "There is currently %d person identified" % len(connected_users) )
            else:
                self.privmsg( nick, "There are currently %d people identified" % len(connected_users) )

#-----------------------------------------------------------------------
# unimplementeds
#-----------------------------------------------------------------------
        elif inputline[:5] == "!info":
            self.privmsg( nick, "Sorry, sir, !saves isn't implemented yet." )
        elif inputline[:6] == "!saves":
            self.privmsg( nick, "Sorry, sir, !saves isn't implemented yet." )
        elif inputline[:6] == "!clear":
            self.privmsg( nick, "Sorry, sir, !clear isn't implemented yet." )

#-----------------------------------------------------------------------
# default
#-----------------------------------------------------------------------
        else:
            self.privmsg( nick, "Sorry, Meldrew doesn't understand that command.  Try !help." )


#-----------------------------------------------------------------------
    def privmsg( self, nick, cmd):
        c = self.connection
        print "<%s>: %s" % (nick, cmd)
        c.privmsg( nick, cmd )
	time.sleep(1)


# MAIN
# main Meldrew processing begins here
def main():
    #import sys
    #import os
    #import string
    #import anydbm

    def load_config():
        print "parsing configuration..."
        config_file = open( "meldrew.rc", "r" )
        config_done = 0;
        adding_to_interpreters = 0;
        adding_to_games = 0;
        while ( config_done == 0 ):
            line = config_file.readline()
            line = line[:-1] # chomp off \n
            if ( line[0:1] != "#" ) & (line != ""):
                if ( line == "INTERPRETERS:" ):
                    adding_to_interpreters = 1;
                    adding_to_games        = 0;
                elif ( line == "GAMES:" ):
                    adding_to_interpreters = 0;
                    adding_to_games        = 1;
                elif ( line == "END" ):
                    config_done = 1;
                elif ( adding_to_interpreters == 1):
                    new_terp = string.splitfields( line, "|" )
                    terp_list.append( new_terp );
                elif ( adding_to_games == 1):
                    new_game = string.splitfields( line, "|" )
                    game_list.append( new_game );
    #end load_config()

    # initialization: load config
    load_config()

    # opening message
    #print "Meldrew reporting for duty, sir!"
    #print ""

    nickname = "meldrew"
    channel = "#meldrew"
    server = "irc.openprojects.net"
    port = 6667

    bot = MeldrewBot( channel, nickname, server, port)
    bot.start()

#-----------------------------------------------------------------------
if __name__ == "__main__":
    main()
